
from __future__ import print_function
import yaml, pprint


my_yml = """

# Detectors
Frelon21 : 
    name : "Frelon21"
    type : "2Ddetector"
    dark : "/data/id11/.../dark1s.edf"
    flat : "/data/id11/.../flat.edf"
    spline : "/data/id11/.../frelon21_mar16.spline"

Frelon4M :
    name : "Frelon4M"
    type : "2Ddetector"
    dark : "/data/id11/.../dark1s.edf"
    flat : "/data/id11/.../flat.edf"
    spline : "/data/id11/.../Frelon4M.spline"

pico0 :
    name : "pico0"
    type : "0Ddetector"
    # e.g. photons to pA current converter
    gain : 1.e7

mca :
    name : "mca"
    type : "1Ddetector"
    nchannels : 4096
    inifile : SDD_53keV.ini
    gain :  #  4096/53.0

# Experiment Stations
3dxrd: 
    - name : 3dxrd
    - wavelength : 0.1234
    - Positioners:
        - omega : 0.0
        - wedge : 0.0
        - samtz : 0.0
    - Detectors:
        - name : Frelon21 
          detx : 120.0  
        - name : Frelon4M
          ffdtz1 : 220.0

Fe3O4: 
    - name : magnetite
    - space_group : Fd-3m
    - composition : { 'Fe':3, 'O':4 }
    - unit_cell : [ 8.39, 8.39, 8.39, 90., 90., 90. ]

Silicon:
    - name : silicon
    - space_group : Fd-3m
    - composition : { 'Si':1 }
    - unit_cell : [ 5.43094,5.430945.43094,90,90,90 ] 

scan1:
    - instrument:
      - 3dxrd
    - sample:
      - Silicon
      - Fe3O4
    - measurement:
      - Frelon21: ["test0001.edf", "test0002.edf", "test0003.edf" ]
      - omega:    [ 0.0 ,          1.0,            2.0]
      - epoch:    [ 0.0 ,          5.0,            10.0]
      - srcur:    [ 190.0 ,      189.9,           189.8]

"""

y = yaml.load(my_yml)
pprint.pprint(y)





from ImageD11 import cImageD11

import unittest
import numpy as np

class test1(unittest.TestCase):
    def test_put(self):
        data = np.zeros(10,np.float32)
        ind  = np.arange(10).astype(np.intp)
        vals = np.ones(10,np.float32)
        cImageD11.put_incr( data, ind, vals )
        assert (data == vals).all()
        cImageD11.put_incr( data, ind, vals )
        assert (data == 2*vals).all()

    def test_put_twice(self):
        data = np.zeros(10,np.float32)
        ind  = np.ones(10,np.intp)
        vals = np.ones(10,np.float32)
        cImageD11.put_incr( data, ind, vals )
        assert (data == np.array( [0, 10] + [0]*8 , np.float)).all()
        cImageD11.put_incr( data, ind, vals )
        assert (data == np.array( [0, 20] + [0]*8 , np.float)).all()

    def test_as_flat(self):
        data = np.zeros( (10, 10), np.float32 )
        ind  = np.ones( 10 , np.intp)*50
        vals = np.ones( 10 , np.float32)
        cImageD11.put_incr( np.ravel(data) , ind, vals )
        assert ( np.ravel(data)[50] == 10 )
        assert ( np.ravel(data)[49] == 0 )
        assert ( np.ravel(data)[51] == 0 )

if __name__=="__main__":
    unittest.main()
        





# Make a project file in a hdf.

# "Scans" ... dimensions ?
#     omega / dty / pz / load / temperature ... etc

# Images ...

# example : 
#
import os
dataRoot = "/data/id11/nanoscope/Commissioning/2018Apr/difftomo_al_quick"
os.chdir( dataRoot )
#dataRoot = "./"
# interlaced = 1
# roundtrip = 1
# nimg = 360
# extensions per folder
imgs = []
for i in range(360):
    imgs.append("%d_%04d.edf"%(0,i))
    imgs.append("%d_%04d.edf"%(1,359-i))

stems = ["Al_y%03d_"%(i) for i in range(-60, 61)]

import h5py, fabio, os

h = h5py.File("/tmp/demo.hdf" , "w" )
for stem in stems:
    g = h.require_group(stem)
    fname = os.path.join( stem, stem+imgs[0] )
    im = fabio.open( fname )
    # hacky
    bsize = im._frames[0].size
    bstart = im._frames[0].start
    bshape = im.data.shape
    # here : verify headers
    extlist = [ (os.path.join( stem, stem+i ), bstart, bsize)
                for i in imgs ]
    g.create_dataset( "images",
                      shape = (len(imgs), bshape[0], bshape[1]),
                      dtype = im.data.dtype,
                      external = extlist )
    h.close()
                      
    
    







from __future__ import print_function

from PyQt4 import Qt
import numpy, sys, time, os

"""
Read an image from a Tango Device
Display it in an opencv window, tracking features

Assumes an RGB picture (=> dim_x is 3*image size)
"""

if 'TANGO_HOST' not in os.environ:
    os.environ['TANGO_HOST'] = "crunch:20000"
print("Using TANGO_HOST = ", os.environ['TANGO_HOST'])

import PyTango

SLEEPER = 1000



def normalise( im ):
    """ Set mean to 0, std to 1 """
    avg = numpy.ravel(im).mean()
    std = numpy.ravel(im).std()
    if std > 0:
        return ( im - avg ) / std
    else:
        return im - avg

def cross_correlate(a2, b2, fftshape):
    """ Make 2D cross correlation image """
    # FFT:
    a2 = numpy.fft.rfft2(a2, s = fftshape)
    b2 = numpy.fft.rfft2(b2, s = fftshape)
    # Cross correlate
    c = numpy.fft.irfft2( a2 * b2.conj() )
    return numpy.fft.fftshift( c )

def find_offset( c ):
    """ Co-ordinates of max in a 2D cross correlation """
    flatmax = c.argmax()
    # print flatmax, c.flat[flatmax]
    dim0 = int(flatmax/c.shape[1])
    dim1 = flatmax - c.shape[1] * dim0
    # print dim0,dim1,c[dim0,dim1]
    roi = c[ dim0-6:dim0+7, dim1-6:dim1+7 ]
    troi = roi.ravel().sum()
    x  = numpy.dot( roi.sum(axis=1), numpy.arange(dim0-6,dim0+7) ) / troi
    y  = numpy.dot( roi.sum(axis=0), numpy.arange(dim1-6,dim1+7) ) / troi
    # Average the local area ?
    return x, y


def register_image( im1, im2 ):
    """ We take the current and place it onto reference """

    ref = normalise(numpy.asarray(im1).sum(axis=2, dtype=numpy.float))
    cur = normalise(numpy.asarray(im2).sum(axis=2, dtype=numpy.float))

    fftshape = ( ref.shape[0] + cur.shape[0] ,
                 ref.shape[1] + cur.shape[1]  )
    cor = cross_correlate( ref, cur, fftshape)
    x, y = find_offset( cor )
    #print x, y, ref.shape, cur.shape
    return x, y







class tangocam:
    def __init__(self, CAMERA_DEVICE = "id11/ccd1394/1"):
        self.cam = PyTango.DeviceProxy(CAMERA_DEVICE)

    def getimage_numpy(self):
        """ return the image as a numeric array """
        res = self.cam.read_attribute_as_str("Image")
        ar = numpy.reshape(numpy.fromstring(res.value,
                                            numpy.uint8) ,
                                   ( res.dim_y , res.dim_x/3 , 3 ))
        return ar


class camreader( Qt.QThread ):
    def __init__(self, lock, cam, im, parent = None):
        super(camreader,self).__init__(parent)
        self.lock = lock
        self.cam = cam
        self.im = im
    def run(self):
        while 1:
            try:
                self.lock.lockForWrite()
                self.im[:,:] = self.cam.getimage_numpy()
            except:
                pass
            self.lock.unlock()
            time.sleep(SLEEPER/1000+0.1)


class ImageQt(Qt.QImage):

    def __init__(self, npyar):
#        start = time.time()
        data = None
        colortable = None
        assert type(npyar) == numpy.ndarray
        assert len(npyar.shape) == 3
        assert npyar.shape[2] == 3
        assert npyar.dtype == numpy.uint8
        h , w = npyar.shape[0:2]
	try: 
            self.__data = npyar.data
	    Qt.QImage.__init__(self, self.__data, w, h, Qt.QImage.Format_RGB888 )
        except:
            im = numpy.zeros( (h, w, 4),
                          numpy.uint8)
            im[:,:,0] = npyar[:,:,2]
            im[:,:,1] = npyar[:,:,1]
            im[:,:,2] = npyar[:,:,0]  
            data = im.tostring()
            self.__data = data or im.tostring()
            Qt.QImage.__init__(self, self.__data, w, h, Qt.QImage.Format_RGB32 )
#        print "Cvt", time.time()-start


class qtcam( Qt.QApplication ):
    def __init__(self, args):

        self.cam = tangocam()
        self.numpyarray = self.cam.getimage_numpy()
        # reference is div/2 + 128 for later subtracting data/2
        self.refdata = self.numpyarray.copy()
        self.refimage = self.numpyarray/2 + 128
        self.last = self.numpyarray.copy()
        self.scalfac = 0.5
        self.iframe = 0

        Qt.QApplication.__init__(self, args )
    
        self.lock = Qt.QReadWriteLock()
        self.camreader = camreader( self.lock, self.cam, self.numpyarray )
        self.camreader.start()

        self.addwidgets()

        self.connect( self.loadbutton,
                Qt.SIGNAL("clicked()"),
                self.loadreference )

        self.connect( self.savebutton,
                Qt.SIGNAL("clicked()"),
                self.savereference )
        
        self.connect( self.zoombutton,
                Qt.SIGNAL("clicked()"),
                self.setscalfac )

        self.connect( self.calcbutton,
                Qt.SIGNAL("clicked()"),
                self.compute )

        self.update()

        self.timer = Qt.QTimer()
        self.timer.setInterval( SLEEPER )
        self.connect( self.timer,
                Qt.SIGNAL("timeout()"),
                self.update )
        self.update()
        self.timer.start()

        self.exec_()


    def addwidgets(self):
        self.form = Qt.QDialog()
        self.form.setWindowTitle( "ID11 Camera" )
        self.vl = Qt.QGridLayout()

        self.vl.addWidget(  Qt.QLabel ("<h2>Live Feed</h2>" ),0,0 )
        self.zoombutton = Qt.QPushButton("Set zoom")
        self.vl.addWidget( self.zoombutton,0,1 )

        self.savebutton = Qt.QPushButton( "Save Current")
        self.vl.addWidget( self.savebutton, 0, 2)

        self.vl.addWidget(  Qt.QLabel ("<h2>Difference</h2>" ),0,3 )

        self.loadbutton = Qt.QPushButton( "Load Reference")
        self.vl.addWidget( self.loadbutton, 0,4)

        self.calcbutton = Qt.QPushButton("Compute")
        self.vl.addWidget( self.calcbutton,0,5 )

        self.camlabel = Qt.QLabel()
        self.vl.addWidget( self.camlabel,1,0,1,3 )

        self.difflabel = Qt.QLabel()
        self.vl.addWidget( self.difflabel,1,3,1,3 )

        self.form.setLayout(self.vl)
        self.form.show()


    def compute(self):
        self.lock.lockForRead()
        if self.numpyarray[0,0,0] != self.last[0,0,0]:
            self.last = self.numpyarray.copy()
        self.lock.unlock()
        x, y = register_image( self.refdata, self.last )
        print("\n", x - self.refdata.shape[0], y - self.refdata.shape[1])
        Qt.QMessageBox.information (self.form, 
                "Computed Offset", 
                "Vertical  = %.5f  \nHorizontal = %.5f"%(
                    x - self.refdata.shape[0],
                    y - self.refdata.shape[1]), 
                Qt.QMessageBox.Ok)


    def setscalfac(self):
        val, ok = Qt.QInputDialog.getDouble(self.form, 
                "Scale factor", 
                "Please enter the image scale",
                self.scalfac, #value = 
                0.01, #minValue = 
                10 , #maxValue 
                 3)       #         int decimals 
        if ok:
            self.scalfac = val
        self.vl.invalidate()
        self.update(Forced=True)

    def loadreference(self):
        fn = Qt.QFileDialog.getOpenFileName(
                self.form,
                "Choose a file to open", 
                os.getcwd()
                )   
        if fn == "":
            return
        ar = numpy.fromstring( open(fn,"rb").read(),
                               numpy.uint8)
        self.refdata = ar.reshape( self.refimage.shape )
        self.refimage = self.refdata/2 + 128
        self.update(Forced=True)

    def savereference(self):
        fn = Qt.QFileDialog.getSaveFileName(
                self.form,
                "Give the filename to save", 
                os.getcwd()
                )   
        if fn == "":
            return
        open( fn, "wb" ).write( self.last.tostring() )
        


    def array_to_display(self, ar):
        i = ImageQt( ar )
        h, w, junk = ar.shape
        return Qt.QPixmap.fromImage(
            Qt.QImage.scaled( i,
                              int(w*self.scalfac),
                              int(h*self.scalfac) ) )

    

    def update(self, Forced=False):
        start = time.time()
        changed = False
        try:
            self.lock.lockForRead()            
            if not (self.numpyarray[0,0,:]==self.last[0,0,:]).all() or Forced:
                self.last = self.numpyarray.copy()
                changed = True
        except:
            print("Got exception")
            pass
        self.lock.unlock()
        if not changed:
            return
        diff = numpy.subtract( self.refimage, self.last/2 )
        pml = self.array_to_display(self.last)
        pmd = self.array_to_display(diff)
        self.camlabel.setPixmap( pml )
        self.difflabel.setPixmap( pmd )
        self.camlabel.show()
        self.difflabel.show()
        self.iframe += 1
        print("Nframes %5d %.5f/s\r"%(self.iframe,
                                      time.time()-start), end=' ')
        sys.stdout.flush()

if __name__=="__main__":
    qtcam(sys.argv)




from __future__ import print_function



# Kindly contributed by Martijn Van Hulzen, Delft, 2016.


# script that increases signal to background ratio by 
# looking for the minimum and maximum value of a pixel in a set of edfs 
# (usually rotated along the z-axis over an omega range) and then subtracting 
# the min from max it also produces the average edf per cycle



import fabio, glob, numpy as np, sys, os

stem, f, l, n = sys.argv[1:] # define base filename (stem), first framenr (f), 
         # last framenr (l) and nr of frames in a cycle (n) as arguments
f=int(f) # turn first framenr into integer
l=int(l) # turn last framenr  into integer
n=int(n) # turn number of frames in a cycle into integer
m = 0    # correction added to the framenr in case a framenr is skipped

# loop over the range of edfs from first to last frame with step n
for i in range(f,l,n):
    fn = "%s%04d.edf"%(stem,i+m)     # set the filename
    # loop until a valid filename is found
    while not os.path.isfile(fn) and m < 10:  
        # print the filename that does not exist
        print("%s does not exist" % fn)        
        # increase correction by one because the frame does not exist 
        m = m + 1                            
        # set the new filename now that m = m + 1 
        fn = "%s%04d.edf"%(stem,i+m)         
    if m > 9:
        print ("Stopping, as too many files do not exist")
        break              # break when too many filenames are missing
    im = fabio.open(fn)            # read the frame data
    s = im.data.astype(np.float32) # assign floating point data to s
                                   # used to determine the average
    lo = s.copy()                  # copy for determining minimum
    hi = s.copy()                  # copy for determining maximum
    for j in range(n):                     # loop over a cycle of frames
        fn = "%s%04d.edf" % (stem,i+j+m)         # set frame filename
        while not os.path.isfile(fn) and m < 10: # check whether filename exists
            # file does not exist, increase m and try again
            m = m + 1                                
            # print filename that does not exist
            print("%s does not exist" % fn)      
            # set new filename      
            fn = "%s%04d.edf" % (stem,i+j+m)    
        if m > 9: 
            print ("Stopping, as too many files do not exist")
            break               # break when too many filenames are missing
        print("%s" % fn)         # print frame to be processed
        f = fabio.open(fn).data # retrieve frame data
        s = s + f               # add new frame to previously added frames
        # determine the min between the current minimum and the current frame
        lo = np.minimum( lo, f )
        # determine the max between the current maximum and the current frame
        hi = np.maximum( hi, f )                
    if m > 9: 
        break                   # break out if files are not found
    s = s / n   # determine the average by dividing by the number of files
    # assign back to original structure that was created at the beginning 
    # of the process (reuse)
    im.data = s  
    im.write( "avg%04d.edf"%(i/n)) # write the average to file
    print("avg%04d.edf"%(i/n))      # print the average filename
    im.data = hi-lo                # assign the max - min to the data structure
    im.write( "pks%04d.edf"%(i/n)) # write the peaks to file
    print("pks%04d.edf"%(i/n))      # print the peaks filename




import os, sys, numpy as np
from ImageD11.columnfile import columnfile


flippers = {
 "_f0" : [[1,0],[0,1]],
 "_f1" : [[-1,0],[0,1]],
 "_f2" : [[1,0],[0,-1]],
 "_f3" : [[-1,0],[0,-1]],
 "_f4" : [[0,1],[1,0]],
 "_f5" : [[0,-1],[1,0]],
 "_f6" : [[0,1],[-1,0]],
 "_f7" : [[0,-1],[-1,0]],
}
c = columnfile( sys.argv[1] )
splinefile = sys.argv[2]
dims = (2048,2048)
sf = c.s_raw.copy(), c.f_raw.copy()
for name in list(flippers.keys()):
    newname = sys.argv[1].replace(".flt",name)+".flt"
    newsf = np.dot( flippers[name], sf )
    origin = np.dot( flippers[name], dims ).clip(-np.inf, 0 )
    c.s_raw[:], c.f_raw[:] = (newsf.T + origin).T
    c.writefile( newname )

    cmd = "fix_spline.py %s %s %s"%(
        newname, newname.replace(".flt","_s.flt"), splinefile ) 
    os.system( cmd )

    # Now try to reverse flip
    d = columnfile( newname.replace(".flt","_s.flt") )
    d.s_raw[:] = sf[0] # copy back original
    d.f_raw[:] = sf[1]
    sfc = np.array( (d.sc.copy(), d.fc.copy()) ) 
    newsfc = np.dot( np.linalg.inv( flippers[name] ) , (sfc.T - origin).T ) 
    d.sc[:], d.fc[:] = newsfc
    d.writefile( newname.replace(".flt","_sfl.flt") )




from __future__ import print_function, division
"""
Determine Bravias lattice for UBI

Return transformation matrix

Due to:
ferraris_and_ivaldi

    Acta Cryst (1983) A39 595-596

    Determine the multiplicities of vectors in a computed powder pattern
    Match these to multiplicities according to lattice types
    ... appeals to powder person
    Tolerance is for grouping vector lengths.
    

"""

import sys
import numpy as np, pylab as pl


from ImageD11.unitcell import unitcell
from ImageD11.grain import read_grain_file
from ImageD11.indexing import ubitocellpars



def fi_triclinic(ubi):
    """ Ensure we have the 3 shortest non-coplanar vectors """
    return ubi

def fi_monoclinic(ubi, cell, mult):
    """ 
    Choose vectors having mult == 2
    These are either along b or in ac plane
    If centred then one mult 4 vector is shorter and cell par
    """
    possibles = []
    for i in range(len(mult)):
        if mult[i] == 2:
            possibles.append( cell.ringhkls(cell.ringds[i])[0] )
    possibles = [cell.ring]
    

def ferraris_and_ivaldi(ubi, tol = 0.001):
    """
    Acta Cryst (1983) A39 595-596

    Determine the multiplicities of vectors in a computed powder pattern
    Match these to multiplicities according to lattice types
    ... appeals to powder person
    Tolerance is for grouping vector lengths.
    """
    cpar = ubitocellpars(ubi)
    cell = unitcell( cp )
    dlim = cell.ds( np.array( [7,7,7] ))
    u.makerings( dlim , tol)
    mult = [ len(cell.ringhkls[d]) for d in cell.ringds ]
    mx = max(mult)
    mn = min(mult)
    cases = {
        (2,  2) : fi_triclinic,
        (4,  2) : fi_monoclinic,
        (8,  2) : fi_orthorhombic,
        (12, 2) : fi_rhombohedral,
        (16, 2) : fi_tetragonal,
        (24, 2) : fi_hexagonal,
        (48, 2) : fi_cubic
    }

def vecang( v1, v2 ):
    cosa = np.dot(v1,v2)/np.sqrt( np.dot( v1, v1 ) * np.dot( v2, v2 ))
    return np.degrees( np.arccos( cosa ))

def lattice_powder( ubi, tol=0.01):
    a,b,c = np.asarray( ubi )
    nmax = 5
    vecs = []
    lens = []
    for ai in range(-nmax,nmax):
        for bi in range(-nmax,nmax):
            for ci in range(-nmax,nmax):
                vec = ai*a + bi*b + ci*c
                modv = np.sqrt( np.dot(vec, vec) )
                lens.append( modv )
                vecs.append( vec )
    order = np.argsort( lens )
    lens  = np.take( lens, order )
    vecs  = np.take( vecs, order ,axis=0 )
    steps = (lens[1:]-lens[:-1])/lens[1:]   # diff divided by upper
    gaps  = steps > tol
    rings = np.cumsum(gaps)
    mults, bins = np.histogram( rings, bins=np.arange(-0.5, rings[-1]+0.5, 1 ) )
    print(rings)
    print(mults)
    nr = min( 10, rings.max() )
    for i in range( nr, 0 , -1 ):
        # this ring:
        M = mults[i]
        print("Powder ring",i,"multiplicity", M )
        print("   Angles between vectors:")
        vl = vecs[ 1: ][ rings==i ]
        vll = lens[ 1: ][ rings==i ]
        for j in range(M):
            print ("%2d:"%(j),vl[j],vll[j])
        print("  ",end="")
        for j in range(M-1,0,-1):
            print("%7d"%(j),end=" ")
        print()
        for j in range(M-1):
            print(" %2d: "%(j), end="")
            for k in range(M-1, j, -1):
                print( "%7.3f"%( vecang( vl[j],vl[k] ) ) , end=" ")
            print()
    print( "Cell pars before",ubitocellpars(ubi))

        
def katayama(ubi):
    """
    J. Appl. Cryst (1986) 19 69-72
    
    Take four vectors a,b,c and d=-(a+b+c)
    Compute the 6 pairwise dot products (ab, ac, ad, bc, bd, cd)
    Apply transformations (up to 12) and recompute
    
    etc
    """
    pass


if __name__=="__main__":
    for grain in read_grain_file( sys.argv[1] ):
        print ("ubi:\n",grain.ubi)
        lattice_powder(grain.ubi)





"""
Add documentation to compiled codes
"""


connectedpixels_doc = """
  nblobs = connectedpixels ( data=numpy.array(data, 2D, np.float32),
                             labels=Numeric.array(blob, 2D, np.int32),
                             threshold=float threshold,
                             verbose=Int verbose )
        data is the image to be labelled
        blob is an array to receive pixel -> blob assignments
        threshold is the value above which a pixel assigned to a blob
        verbose flag : 0=quiet, higher = louder printing on stdout
"""

blobproperties_doc = """
     res = blobproperties ( numpy.array(data, 2D, np.float32),
                            numpy.array(blob, 2D, np.int32)  ,
                            int32 npk, 
                            float32 omega,
                            int32 verbose )
        Computes various properties of blobs in a data image
                (created by connectedpixels or other labelling)
        data  = image data
        blob  = integer peak assignments (e.g. connectedpixels)
        omega = third dimension value (which image is this)
        npk    = number of peaks to treat
        verbose flag : 0=quiet, higher = louder printing on stdout
     Returns res : array of peak properties
"""

bloboverlaps_doc = """
 success = bloboverlaps ( numpy.array(blob1, 2D, np.int32), n1  , res1,
                          numpy.array(blob2, 2D, np.int32), n2  , res2,
                          verbose=0)
      merges the results from blob1/res1 into blob2/res2
      blob1 would be the previous image from the series, with its results
        if it does not overlap it will stay untouched in res
        if it overlaps with next image it is passed into that ones res
"""


blob_moments_doc = """
       blob_moments(numpy.array(res1), npk)
           Loop over array filling out moments from sums 
"""



closest_doc = """
    *    From python: closest(cosines_allowed, cosine_measured)
    *    Take 1 peak with hkl's assigned, compute the cosines to
    *    another powder ring (angles between this peak and peaks in
    *    that ring). This extension finds the best peak in the
    *    other ring to pair up with.
    *
    *    More generally closest(array, values) finds the closest
    *    item in values to one of the values in array.  Returns the
    *    difference and the index of the closest thing it finds.
    *
    *    Both arguments should be one dimensional Numeric arrays
    *    of type Numeric.Float
"""

score_doc = """
         From python: score(ubi, gv, tol) where ubi is an orientation
    *    matrix and gv are an array of g-vectors. Returns the number
    *    of g-vectors which have integer hkl peaks within tolerance
    *    tol. Uses the conv_double_to_int_fast function in here for 
    *    factor of !EIGHT! speed increase compared to rounding in 
    *    C. In fact this gives the nearest even integer, instead
    *    of the nearest integer, but we don't care, as a peak having
    *    hkl of 0.5 is nowhere near being indexed anyway.
    *
    *    Returns and integer - number of peaks indexed
    *    UBI is a 3x3 Numeric.Float array (figure out the transposing yourself)
    *    GV is a nx3 Numeric.Float array, and you should try to make the 3 
    *       be the fast index for best performance
"""

score_and_refine_doc = """
    *    score_and_refine(ubi, gv, tol) 
    *    From python: same as score, I hope, but ubi is overwritten
    *    with refined matrix following paciorek algorithm which is 
    *    in indexing.py
"""

score_and_assign_doc = """    
    *    score_and_assign( ubi, gv, tol, drlv, labels, (int) label)
    *    as for score, but assignments are only made if drlv is lower than
    *    the previous. Fills in label with the new label
"""

refine_assigned_doc = """
    * 5. refine_assigned( ubi, gv, labels, label, weight) 
    *    Use only the peaks where label matches labels in refinement 
    *    ...following again the Paciorek algorithm 
"""

put_incr_doc = """
    * 6. put_incr( data, indices, values)
    *    pretty much as numeric.put but as increments
"""

weighted_refine_doc = """
    * 7. weighted_refine(ubi, gv, tol, weights) 
    *    From python: same as score, but with weights, and ubi is overwritten
    *    with refined matrix following paciorek algorithm which is 
    *    in indexing.py
"""




from __future__ import print_function

"""
Row/column array type

Inherits from numpy.ndarray

To be used for lattice reduction to apply to g-vectors and patterson
peaks equally well and in a coherent way.
"""

from numpy import dot, round_, array, float, allclose, asarray, fabs,\
    argmin, argmax, sqrt, argsort, take, sum, where, ndarray, eye,\
    zeros, cross
from numpy.linalg import inv, LinAlgError

# Confirm that dot'ting a 3x3 matrix with a 3x10 gives a 3x10
assert dot(eye(3), zeros( (3, 10) ) ).shape == (3, 10), \
    "Numpy dot insanity problem"
           
# It is unclear why it is not a 10x3 result (row/col vectors)

try:
    dot(eye(3), zeros( (10, 3) ) )
    raise Exception("Numpy dot insanity problem")
except ValueError: 
    pass
except:
    print("Unexpected exception when checking numpy behaviour")
    raise



# Based on http://www.scipy.org/Subclasses
class rc_array(ndarray):
    """ 
    Row/Column array
    Represent a list of row or column vectors
    """
    def __new__(subtype, data, direction=None, dtype=None, copy=False):
        """ 
        Mostly as from example
        direction is one of row / column
        """
        subarr = array(data, dtype=dtype, copy=copy)
        subarr = subarr.view(subtype)
        if direction is not None:
            subarr.direction = direction
        elif hasattr(data, 'info'):
            subarr.direction = data.direction
        return subarr
    
    def __array_finalize__(self, obj):
        """ 
        Fill in a default row arg to direction
        self/obj??
        """
        self.direction = getattr(obj, 'direction', 'row' )
        

    def __str__(self):
        """ Used for printing """
        desc = \
"""{%(data)s in %(direction)s direction}"""
        return desc % { 'data': super.__str__(self), 
                        'direction' : self.direction }

    def __iter__(self):
        """
        Iterate depending on rows columns
        Use to get [ v for v in rr_array ]
        """
        # print "iter called"
        if self.direction == 'row':
            return ndarray.__iter__(self)
        elif self.direction == 'col':
            return ndarray.__iter__(self.T)
        else:
            raise Exception("rc_array with direction not in row|col")

    def norm2(self):
        """ sum(v*v,axis=? for row or col """
        return sum( self*self, axis=self.vector_axis())

    def vector_axis(self):
        """ The axis which has the 3 on it """
        return ['col', 'row'].index(self.direction)

    def nb_vector_axis(self):
        """ The axis which has the n on it """
        return ['row', 'col'].index(self.direction)

    def other_direction(self):
        """ The one which is not self.direction
        """
        return ['row', 'col'][self.vector_axis()]

    def check(self):
        """ 
        Ensure we have an rc_array which is well behaved 
        Pattern assert(check(v)) should disappear in optimiser
        """
        assert hasattr(self, 'direction')
        assert self.direction in ['row', 'col']
        if len(self.shape) == 1:
            assert self.shape[0] == 3, str(self)
        elif len(self.shape) == 2:
            if self.direction == 'row':
                assert self.shape[1] == 3
            if self.direction == 'col':
                assert self.shape[0] == 3
        else:
            raise Exception("Only 1D or 2D rc_arrays allowed so far")
        return True

    def nvectors(self):
        return self.shape[self.nb_vector_axis()]

    def flip(self, mat):
        """
        Flip row to col or col.row
        """
        assert self.check()
        assert mat.shape == (3, 3)
        if self.direction == 'row':
            ret = dot( mat , self.T)
        if self.direction == 'col':
            ret = dot( mat, self).T
        ret = rc_array( ret, direction = self.other_direction())
        ret.check()
        assert ret.shape == self.shape[::-1],"Shape mismatch in flip"
        return ret

    def inv(self):
        """
        Inverse matrix of self
        """
        assert self.shape == (3,3)
        ret = inv(self)
        return rc_array(ret, self.other_direction())

if __name__=="__main__":


    v = rc_array([1,2,3],direction='row')
    assert v.other_direction() == 'col'

    v = rc_array([[1,2,3],[4,5,6]],direction='row')

    assert v.other_direction() == 'col'
    assert v.flip(eye(3)).direction == 'col'
    assert allclose( v.flip(eye(3)) , v.T )
    assert v.norm2().shape == ( v.shape[ v.nb_vector_axis() ] ,)
    assert allclose( v.norm2(), array([ 1+4+9, 16+25+36 ])), str(v)
    

    v = rc_array(array( [[1,2,3],[4,5,6]] ).T ,direction='col')

    assert v.other_direction() == 'row'
    assert v.flip(eye(3)).direction == 'row'
    assert allclose( v.flip(eye(3)) , v.T )

    assert v.norm2().shape == ( v.shape[ v.nb_vector_axis() ] ,)

    assert allclose( v.norm2(), array([ 1+4+9, 16+25+36 ])), str(v)




import numpy, ImageD11.cImageD11
numpy.random.seed(42)
im = numpy.random.random((2048,2048)).astype(numpy.float32)
print ImageD11.cImageD11.clip_std(im.ravel(),1)
print im.mean(),im.std(),im.var()
ImageD11.cImageD11.clip_std(im.ravel(),1)




import os
n = 0
i = -12
j = -12
nline = 0
while i < 12.1:
    f = open("%d.sh"%(nline),"w")
    f.write("#!/bin/bash\n")
    while j < 12.1:
        f.write( "./idx_point.sh %f %f %d\n"%(i,j,n))
        n += 1
        j += 0.5
    i += 0.5
    
    j = -12
    f.close()
    os.system("chmod a+x %d.sh"%(nline))
    os.system("oarsub ./%d.sh"%(nline))
    nline += 1
